"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var electron_1 = require("electron");
var eventHandlingDelay = 100;
function isNumber(num) {
    return typeof num === "number" && !isNaN(num);
}
var ElectronWindowState = /** @class */ (function () {
    function ElectronWindowState(config, state, onSave) {
        this.state = {};
        this.onSave = onSave;
        this.stateChangeHandler = this.stateChangeHandler.bind(this);
        this.closeHandler = this.closeHandler.bind(this);
        this.closedHandler = this.closedHandler.bind(this);
        this.loadState(state, config || {});
    }
    ElectronWindowState.prototype.loadState = function (state, config) {
        this.state = state;
        this.normState(this.state, config);
        // Check state validity
        this.validateState(this.state);
    };
    ElectronWindowState.prototype.normState = function (state, config) {
        if (!isNumber(state.x)) {
            state.x = config.defaultX || 0;
        }
        state.x = Math.floor(state.x);
        if (!isNumber(state.y)) {
            state.y = config.defaultY || 0;
        }
        state.y = Math.floor(state.y);
        if (!isNumber(state.width)) {
            state.width = config.defaultWidth || 800;
        }
        state.width = Math.floor(state.width);
        if (!isNumber(state.height)) {
            state.height = config.defaultHeight || 600;
        }
        state.height = Math.floor(state.height);
        if (state.useContentSize == null) {
            state.useContentSize = config.useContentSize || false;
        }
    };
    ElectronWindowState.prototype.isNormal = function (win) {
        return !win.isMaximized() && !win.isMinimized() && !win.isFullScreen();
    };
    ElectronWindowState.prototype.validateState = function (state) {
        // Get the display where the window was last open
        var displayBounds = electron_1.screen.getDisplayMatching(state).bounds;
        if (state.y >= displayBounds.y &&
            state.y + state.height <= displayBounds.y + displayBounds.height &&
            state.x >= displayBounds.x &&
            state.x + state.width <= displayBounds.x + displayBounds.width) {
            return;
        }
        state.width = Math.min(displayBounds.width, state.width);
        state.height = Math.min(displayBounds.height, state.height);
        state.x = Math.floor(displayBounds.x + (displayBounds.width - state.width) / 2);
        state.y = Math.floor(displayBounds.y + (displayBounds.height - state.height) / 2);
    };
    ElectronWindowState.prototype.updateState = function () {
        var state = this.state;
        var win = this.winRef;
        // don't throw an error when window was closed
        try {
            if (this.isNormal(win)) {
                var winBounds = state.useContentSize ? win.getContentBounds() : win.getBounds();
                state.x = winBounds.x;
                state.y = winBounds.y;
                state.width = winBounds.width;
                state.height = winBounds.height;
            }
            state.isMaximized = win.isMaximized();
            state.isFullScreen = win.isFullScreen();
        }
        catch (err) {
            console.error(err);
        }
    };
    ElectronWindowState.prototype.stateChangeHandler = function () {
        var _this = this;
        // Handles both 'resize' and 'move'
        clearTimeout(this.stateChangeTimer);
        this.stateChangeTimer = setTimeout(function () { return _this.updateState(); }, eventHandlingDelay);
    };
    ElectronWindowState.prototype.closeHandler = function () {
        this.updateState();
    };
    ElectronWindowState.prototype.closedHandler = function () {
        // Unregister listeners and save state
        this.unmanage();
        this.onSave && this.onSave(this.state);
    };
    ElectronWindowState.prototype.manage = function (win) {
        var state = this.state;
        if (state.isMaximized) {
            win.maximize();
        }
        if (state.isFullScreen) {
            win.setFullScreen(true);
        }
        win.on("resize", this.stateChangeHandler);
        win.on("move", this.stateChangeHandler);
        win.on("close", this.closeHandler);
        win.on("closed", this.closedHandler);
        this.winRef = win;
    };
    ElectronWindowState.prototype.unmanage = function () {
        var winRef = this.winRef;
        if (winRef) {
            winRef.removeListener("resize", this.stateChangeHandler);
            winRef.removeListener("move", this.stateChangeHandler);
            clearTimeout(this.stateChangeTimer);
            winRef.removeListener("close", this.closeHandler);
            winRef.removeListener("closed", this.closedHandler);
            this.winRef = null;
        }
    };
    return ElectronWindowState;
}());
exports.default = ElectronWindowState;
